import paho.mqtt.client as PahoMQTT
import json
import subprocess
import random
import logging
from logging.handlers import RotatingFileHandler
import threading
import time
import ast
import can
from can.interfaces import socketcan
import struct
import os
import gzip
import queue


version = "1.3.1"
confFilename = "/etc/realtime/MQTT-conf_realtime.json"
logFilename = "/var/log/realtime.log"

def CANFactory(RTStart, CANStart, canIf,msg_queue):
	'''This threading function initializes the CAN bus.
	When it starts it takes the data filter and it appends the messages in the queue if the can 
	streaming is enabled via CANStart event.	
	'''
	global can_filters
	global buffer

	logging.info("Started thread")

	while True:
		#wait realtime enabling
		logging.debug("Waiting RTStart")
		RTStart.wait()

		#open the can bus interface
		with can.interfaces.socketcan.SocketcanBus(channel=canIf, receive_own_messages=False, fd=False) as bus:
			logging.info("SocketCAN open on channel: {}".format(canIf))
			#logging.debug(type(can_filters["can_filters"][0]))
			bus.set_filters(can_filters["can_filters"])		
			logging.debug("Start CAN listeners with filters:\n{}".format(can_filters["can_filters"]))

			ext_filters = can_filters["ext_filters"]
			logging.debug("ext_filters: {}".format(ext_filters))
			
			#initalize the buffer list
			buffer = []

			#initialize the variables to None
			for var in ext_filters["variables"]:
				buffer.append((None,0))	#in each position insert a tuple with the value of the variable and the valid flag
			
			logging.info("Buffer initialized")
			#till the realtime is enabled extract new messages
			while RTStart.is_set():
				msg = bus.recv(timeout=2.0)
				#logging.debug("Message receive from CAN")

				#if the CAN streaming is valid put the incoming message into the queue
				if msg and CANStart.is_set():
					msg_queue.put(msg)
					#logging.debug("Message append in listener_buffer queue")
			#logging.info("Stop receiving messages, ready to close socketcan")




def BufferFactory (msg_queue, CANStart, RTStart, bufferLock):
	'''Threading function that wait untill the can streaming is enabled via CANStart event.
	It takes the messages form the queue and it prepares the buffer with the raw varibales.
	'''
	global buffer
	global can_filters


	logging.info("Started thread")
	while True:
		CANStart.wait()
		
		#receive msg
		msg = msg_queue.get()
		#logging.debug("Msg received:\t{}\t{}".format(hex(msg.arbitration_id), msg.data))
		#logging.info("Queue size: {}".format(msg_queue.qsize()))
		#convert the bytearray received from the can into Big-endian hexadecimal string
		if RTStart.is_set():
			unpack_data = format(struct.unpack('<Q', msg.data)[0], '0>16X')
			#logging.debug("unpack data        <{}>".format(unpack_data))

			pgn = msg.arbitration_id 
			#logging.debug("msg arbitration id <{}>".format(pgn))

			for var in can_filters["ext_filters"]["variables"]:
				#check if the pgn collected matches with any variable configured canid
				if var["can_id"] == (pgn & var["can_mask"]):
					#logging.debug("pgn matched with <{}>".format(var["can_id"]))

					if var ["mux"]:
						mux = int(format(int(unpack_data, base=16), '0>64b')[56:], base=2) 
						logging.debug("Mux value extracted: {}\tvar['mux']: {}".format(mux,int(var["mux"])))
						if int(var["mux"]) == mux:
							in_data = int(format(int(unpack_data, base=16), '0>64b')[-var["toBit"]-1:None if not var["fromBit"] else -var["fromBit"]], base=2)
							if in_data > int(var["max"]) or in_data < int(var["min"]):
								if buffer[can_filters["ext_filters"]["variables"].index(var)][1]:
									data = buffer[can_filters["ext_filters"]["variables"].index(var)]
								else:
									data = (in_data,0)
							else:
								data = (in_data,1)
						else:
							data = buffer[can_filters["ext_filters"]["variables"].index(var)]

					elif var["signed"]: 			#SIGNED VARIABLES
						#convert the variable in binary format, then extracts bits and convert in integer
						in_data = int.from_bytes(int(format(int(unpack_data, base=16), '0>64b')[-var["toBit"]-1:None if not var["fromBit"] else -var["fromBit"]], base=2).to_bytes(2,byteorder='big', signed=False), byteorder='big', signed=True)
						if in_data > int(var["max"]) or in_data < int(var["min"]):
							if buffer[can_filters["ext_filters"]["variables"].index(var)][1]:
								data = buffer[can_filters["ext_filters"]["variables"].index(var)]
							else:
								data = (in_data,0)
						else:
							data = (in_data,1)
					elif var["multiple"]:		#MULTIPLE SOURCES VARIABLES
						#convert the variable in binary format, then extracts bits and convert in integer
						in_data = int(format(int(unpack_data, base=16), '0>64b')[-var["toBit"]-1:None if not var["fromBit"] else -var["fromBit"]], base=2)
						logging.debug("Data before switch case <{}>".format(in_data))
						logging.debug("Buffer before assignement <{}>".format(buffer))
						
						if in_data == 0:
							in_data = format((int(buffer[can_filters["ext_filters"]["variables"].index(var)][0], base=2) if buffer[can_filters["ext_filters"]["variables"].index(var)][0] else 0) | 1, '0>16b')
						elif in_data == 1:
							in_data = format((int(buffer[can_filters["ext_filters"]["variables"].index(var)][0],base=2) if buffer[can_filters["ext_filters"]["variables"].index(var)][0] else 0) | 2, '0>16b')
						elif in_data == 2:
							in_data = format((int(buffer[can_filters["ext_filters"]["variables"].index(var)][0], base=2) if buffer[can_filters["ext_filters"]["variables"].index(var)][0] else 0) | 4, '0>16b')
						elif in_data == 3:
							in_data = format((int(buffer[can_filters["ext_filters"]["variables"].index(var)][0], base=2) if buffer[can_filters["ext_filters"]["variables"].index(var)][0] else 0) | 8, '0>16b')
						elif in_data == 4:
							in_data = format((int(buffer[can_filters["ext_filters"]["variables"].index(var)][0],base=2) if buffer[can_filters["ext_filters"]["variables"].index(var)][0] else 0) | 16, '0>16b')
						elif in_data == 5:
							in_data = format((int(buffer[can_filters["ext_filters"]["variables"].index(var)][0],base=2) if buffer[can_filters["ext_filters"]["variables"].index(var)][0] else 0) | 32, '0>16b')
						elif in_data == 6:
							in_data = format((int(buffer[can_filters["ext_filters"]["variables"].index(var)][0],base=2) if buffer[can_filters["ext_filters"]["variables"].index(var)][0] else 0) | 64, '0>16b')
						elif in_data == 7:
							in_data = format((int(buffer[can_filters["ext_filters"]["variables"].index(var)][0],base=2) if buffer[can_filters["ext_filters"]["variables"].index(var)][0] else 0) | 128, '0>16b')
						elif in_data == 8:
							in_data = format((int(buffer[can_filters["ext_filters"]["variables"].index(var)][0],base=2) if buffer[can_filters["ext_filters"]["variables"].index(var)][0] else 0) | 256, '0>16b')
						elif in_data == 9:
							in_data = format((int(buffer[can_filters["ext_filters"]["variables"].index(var)][0],base=2) if buffer[can_filters["ext_filters"]["variables"].index(var)][0] else 0) | 512, '0>16b')
						elif in_data == 10:
							in_data = format((int(buffer[can_filters["ext_filters"]["variables"].index(var)][0],base=2) if buffer[can_filters["ext_filters"]["variables"].index(var)][0] else 0) | 1024, '0>16b')
						elif in_data == 11:
							in_data = format((int(buffer[can_filters["ext_filters"]["variables"].index(var)][0],base=2) if buffer[can_filters["ext_filters"]["variables"].index(var)][0] else 0) | 2048, '0>16b')
						elif in_data == 12:
							in_data = format((int(buffer[can_filters["ext_filters"]["variables"].index(var)][0],base=2) if buffer[can_filters["ext_filters"]["variables"].index(var)][0] else 0) | 4096, '0>16b')
						elif in_data == 254:
							in_data = format((int(buffer[can_filters["ext_filters"]["variables"].index(var)][0],base=2) if buffer[can_filters["ext_filters"]["variables"].index(var)][0] else 0) | 16384, '0>16b')
						elif in_data == 255:
							in_data = format((int(buffer[can_filters["ext_filters"]["variables"].index(var)][0],base=2) if buffer[can_filters["ext_filters"]["variables"].index(var)][0] else 0) | 32768, '0>16b')	
						else:
							if buffer[can_filters["ext_filters"]["variables"].index(var)][0]:	#default case takes the last value
								in_data = format(int(buffer[can_filters["ext_filters"]["variables"].index(var)][0],base=2), '0>16b')
						
						data = (in_data,1)
					
					else:						#UNSIGNED VARIABLES
						#convert the variable in binary format, then extracts bits and convert in integer
						in_data = int(format(int(unpack_data, base=16), '0>64b')[-var["toBit"]-1:None if not var["fromBit"] else -var["fromBit"]], base=2)
						if in_data > int(var["max"]) or in_data < int(var["min"]):
							if buffer[can_filters["ext_filters"]["variables"].index(var)][1]:
								data = buffer[can_filters["ext_filters"]["variables"].index(var)]
							else:
								data = (in_data,0)
						else:
							data = (in_data,1)

					logging.debug("data extracted <{}>".format(data))
					#logging.debug("variable: {}".format(var))
					#logging.debug("pgn: {}".format(pgn))
					#insert the data into the buffer
					bufferLock.acquire()
					logging.debug("BufferLock acquired")
					buffer[can_filters["ext_filters"]["variables"].index(var)] = data
					bufferLock.release()
					#logging.debug("buffer updated <{}>".format(buffer))


def periodicPublisher(tablesMS,RTStart, CANStart, publisher, bufferLock, interval=2.0):
	'''Thread function that periodically tranfroms the variables inside the buffer
	into human-readable data and publishes them with the correct topic
	'''

	global buffer
	global can_filters


	logging.info("Started thread")

	time.sleep(2)
	
	while True:
		RTStart.wait()
		payload_out = {"error": 0,
									"values":[] }
		
		#if the realtime streming is enabled AND the can streaming is enabled the data are published periodically
		while RTStart.is_set() and CANStart.is_set():
			#time to sleep between different publications
			time.sleep(interval)
			logging.info("Start to extract data from buffer")
			payload_out["values"] = []
			#initialize the dictionary to be published
			for var in can_filters["ext_filters"]["variables"]:
				payload_out["values"].append({"id" : var["name"], "value": "NaN"})
			
			logging.debug("payload_out after init:\n{}".format(payload_out))
			logging.debug("len(buffer) <{}>".format(len(buffer)))
			logging.debug("buffer <{}>".format(buffer))

			#convert the raw data into human-readable data
			for i in range(len(buffer)):
				data = 0
				if buffer[i][0] != None:
					raw_data = buffer[i][0]
					logging.debug("i-cycle <{}>\t raw_data int <{}>".format(i,raw_data))

					if can_filters["ext_filters"]["variables"][i]["multiple"]:
						for elem in can_filters["ext_filters"]["variables"][i]["labels"]:
							if elem["val"] == raw_data:
								data = elem["label"]
						for elem in tablesMS[can_filters["ext_filters"]["variables"][i]["multiple"]]:
							logging.debug("Elem in table <{}>".format(elem))
							logging.debug("data[elem] = {}".format(raw_data[elem]))
							if raw_data[elem] == '1':
								data = str(15 - elem)
								break

					else:					
						if buffer[i][1]:
							for elem in can_filters["ext_filters"]["variables"][i]["labels"]:
								if elem["val"] == raw_data:
									data = elem["label"]
							if not data:
								data = raw_data * can_filters["ext_filters"]["variables"][i]["factor"]
								#logging.debug("data int *= <{}>".format(data))
								data += can_filters["ext_filters"]["variables"][i]["offset"]
								#logging.debug("data int += <{}>".format(data))
								data = str(round(data,3))
							
						else:
							data = "NaN"
							for elem in can_filters["ext_filters"]["variables"][i]["labels"]:
								if elem["val"] == raw_data:
									data = elem["label"]

					payload_out["values"][i]["value"] = data
					logging.debug("Payload_out update <{}>".format(payload_out))
			
			payload_out["error"] = 0
			#publish the results
			publisher.Publish(json.dumps(payload_out,separators=(',',':')))
			logging.info("Queue size: {}".format(msg_queue.qsize()))

			bufferLock.acquire()
			logging.debug("BufferLock acquired")
			for i in range(len(buffer)):
				buffer[i] = (None,0)
			bufferLock.release()

		#if the can streaming is NOT enabled it only publishes the key_on status periodically
		time.sleep(interval)
		publisher.Publish('{"error":%d}'%(1 if RTStart.is_set() else 2))
		logging.info("Realtime streaming stopped")

def keyOnManager (CANStart):
	'''Thread function that checks in background the key status and enables or disables the can streaming.
	If the key is switched off the can streaming has to be ignored.
	'''

	logging.info("Started thread")

	while True:
		#get the status of the key
		req_getKey = subprocess.Popen("udpctrl in-get 0", stderr=subprocess.PIPE, shell=True)

		try :
			getKey = 1 if req_getKey.stderr.read().decode('utf-8').split("=")[1][:-1] == ' on' else 0
		except:
			logging.warning("Impossible to get the key status")
			getKey = 0
		
		#controls the status of the key now and the previous registered state
		if not getKey:
			#if now the key is switched off AND it was switched on before
			if CANStart.is_set():
				CANStart.clear()
				logging.warning("Stop CAN acquisition due to key switch OFF")
		else:
			#if now the key is switched on now AND it was switched off before
			if not CANStart.is_set():
				CANStart.set()
				logging.warning("Start CAN acquisition due to key switch ON")

		#time of polling. Fine tune this parameter to be more or less reactive to state changes
		time.sleep(0.5)

 
class MQTTService:
		def __init__(self, init_conf, name, eventStartRT, logging, version):			

			self._paho_mqtt=PahoMQTT.Client(client_id=name,clean_session=False)
			
			#callbacks definition
			self._paho_mqtt.on_connect=self.OnConnect
			self._paho_mqtt.on_message=self.OnMessageReceived
			self._paho_mqtt.on_publish=self.OnPublish
			self._paho_mqtt.on_subscribe=self.OnSubscribe
			self._paho_mqtt.on_disconnect=self.OnDisconnect
			self._paho_mqtt.on_pre_connect=self.OnPreConnect

			self.logging=logging
			self.broker=init_conf["mqtt"]["brokerID"]
			self.port=init_conf["mqtt"]["brokerPort"]
			self.keepalive=init_conf["mqtt"]["keepalive"]

			#topics definition
			self.LTW_topic= name + init_conf["mqtt"]["LTW_topic"]
			self.sub_topic= name + init_conf["mqtt"]["sub_topic"]
			self.pub_topic= name + init_conf["mqtt"]["pub_topic"]

			#subscibed topics callbacks
			self.call_set = name + init_conf["mqtt"]["call_set"]
			self.call_upgrade = name + init_conf["mqtt"]["call_upgrade"]
			self.call_offline = name + init_conf["mqtt"]["call_offline"]
			self.call_version = name + init_conf["mqtt"]["call_version"]

			self.update_url = init_conf["update"]["domain"]

			self.version_MQTT = version
			self.version_conf = init_conf["version"]

			#time to live of the realtime service
			self.realtimeTTL = init_conf["realtimeTTL"]

			#message received callbacks
			self._paho_mqtt.message_callback_add(sub=self.call_set, callback=self.OnSetStream)
			self._paho_mqtt.message_callback_add(sub=self.call_offline, callback=self.OnStopStream)
			self._paho_mqtt.message_callback_add(sub=self.call_upgrade, callback=self.OnUpdate)
			self._paho_mqtt.message_callback_add(sub=self.call_version, callback=self.OnGetVersion)

			self.name=name
			self.RTStart=eventStartRT

			self.unconnect_flag=True
			
			#timer to manage the auto stop after ttl expire
			self.timerStop = threading.Timer(interval=self.realtimeTTL, function=self.autoStopRT)

			
		def start (self):
			# testament and will
			self._paho_mqtt.will_set(topic=self.LTW_topic,qos=2)

			#set reconnection delay between retry
			self._paho_mqtt.reconnect_delay_set(min_delay=1, max_delay=32)

			exception = True
			#manage connection to broker
			while self.unconnect_flag:
				if exception:
					try:
						self._paho_mqtt.connect(host=self.broker, port=self.port, keepalive=self.keepalive)
						logging.info("Connect call")
					except Exception as err:
						logging.warning("Exception occurred: <{}>".format(err))
					else:	
						self._paho_mqtt.loop_start()
						exception = False
						
				self.logging.info("Connection waiting loop")
				time.sleep(2)

 
			# subscribe for a topic
			self._paho_mqtt.subscribe(topic=self.sub_topic,qos=2)	
			#the QoS=2 ensures that messageas with high importamce to be transmitted are surely received by this client.
			#Since the session is not cleared by the borker (CleanSession=False), even if the client is temporarly disconnected, the messages
			#published with QoS=2 are delivered exaclty once when this client reconnects. This is useful with the offline communication.

						

		def stop (self):
			#close the MQTT service
			self._paho_mqtt.unsubscribe(self.sub_topic)
			self._paho_mqtt.loop_stop()
			self._paho_mqtt.disconnect()

		def OnPreConnect (self, *args):
			pass

		def OnConnect (self, paho_mqtt, userdata, flags, rc):
			self.logging.info("%s connected to %s with result code: %d" % (self.name,self.broker, rc))
			#print ("Subscriber connected to %s with result code: %d" % (self.broker, rc))

			if rc == 0:
				self.unconnect_flag = False
			elif rc == 1:
				self.logging.warning("Connection refused due to incorrect protocol version")
			elif rc == 2:
				self.logging.warning("Connection refused due to invalid client identifier")
			elif rc == 3:
				self.logging.warning("Connection refused due to server unavailable")
			elif rc == 4:
				self.logging.warning("Connection refused due to bad username or password")
			elif rc == 5:
				self.logging.warning("Connection refused due to not authorised")
			else:
				self.logging.warning("connection refused due to unknown reason")

		def OnDisconnect (self, paho_mqtt, userdata, rc):
			self.logging.info("Client disconnected with reason <{}>".format(rc))
			self.unconnect_flag = True

		def OnMessageReceived (self, paho_mqtt , userdata, msg):
			# A new message is received
			self.logging.info("Unknown message received")
			self.logging.debug("Topic:<" + msg.topic+">, QoS: <"+str(msg.qos)+"> Message: <"+str(msg.payload) + ">")
			#print ("Topic:'" + msg.topic+"', QoS: '"+str(msg.qos)+"' Message: '"+str(msg.payload) + "'")
			#global output
			#self.Publish(data=msg.topic, topic=self.pub_topic + "message_unknown")

		def OnSetStream (self, paho_mqtt , userdata, msg):
			'''Callback function when a new realtime stream setting is arrived.
			It extracts the dictionary of settings and the set of the Lock event
			for other thread function waiting to start
			'''

			self.logging.info("Setting new CAN filters")
			self.logging.debug("Topic:<" + msg.topic+">, QoS: <"+str(msg.qos)+"> Message: <"+str(msg.payload) + ">")

			global can_filters
			
			#extract filters
			can_filters=json.loads(msg.payload.decode())

			#add can_mask in each variable
			for variable in can_filters["ext_filters"]["variables"]:
				for filtr in can_filters["can_filters"]:
					if variable["can_id"] == filtr["can_id"]:
						variable["can_mask"] = filtr["can_mask"]

			#start CAN factory
			self.RTStart.set()
			logging.warning("RT streaming ENABLED")
			
			#renew the timer if is already running
			if self.timerStop.is_alive():
				self.logging.info("Timer stop already alive. Kill it!")
				self.timerStop.cancel()

			self.timerStop = threading.Timer(interval=self.realtimeTTL, function=self.autoStopRT)
			self.timerStop.start()
			self.logging.info("Timer armed with interval={} sec".format(self.realtimeTTL))

		def OnStopStream (self, paho_mqtt, userdata, msg):
			'''Callback function when the client listening for the realtime stream 
			disconnects ungracefully and so the stream has to be stop
			'''

			self.RTStart.clear()			
			self.logging.warning("RT streaming DISABLED")

			#kill the timer if it is running
			if self.timerStop.is_alive():
				self.logging.info("Timer stop already alive. Kill it!")
				self.timerStop.cancel()

			
		def OnUpdate (self, paho_mqtt, userdata, msg):
			'''Callback function when the application needs to be updated. It launches the automatic update scripting
			that download the application and restart the service
			'''

			#update the MQTT_realtime.py script
			if msg.payload == b'update MQTT service':
				self.logging.info("Updating client...")

				self.logging.info("Launching autoupdate script...")
				forked_child = subprocess.Popen("sh /etc/realtime/MQTT_update.sh " +  self.update_url, shell=True)
				forked_child.wait()
				self.logging.info("Return code from subprocess <{}>".format(forked_child.returncode))

				#self.logging.debug("answer from subprocess {}".format(req.stderr.read().decode('utf-8')))
			elif msg.payload == b'update script':
				#update the script update
				self.logging.info("Updating script")
				_ = subprocess.Popen("sh /etc/realtime/script_update.sh " + self.update_url, shell=True)

			elif msg.payload == b'update conf':
				#update the configuration file and restart the service
				self.logging.info("Updating conf file...")
				forked_child = subprocess.Popen("sh /etc/realtime/MQTT_update.sh " +  self.update_url + " conf", shell=True)
				forked_child.wait()
				self.logging.info("Return code from subprocess <{}>".format(forked_child.returncode))

		def OnGetVersion (self, paho_mqtt, userdata, msg):
			'''Callback function when someone requires the version of the python script and the conf file'''
			self.logging.info("version_MQTT: " + self.version_MQTT + ", version_conf: " + self.version_conf)
			data = '{"version_MQTT":' + self.version_MQTT + ',"version_conf":' + self.version_conf +"}"
			self.Publish(data,self.name + "/versions/")


		def Publish(self, data, topic=0):		
			self._paho_mqtt.publish(self.pub_topic if topic == 0 else topic, data)   

		def OnPublish(self, paho_mqtt, userdata, mid):
			#A new data is published
			self.logging.info("Published new data")
			#logging.debug("Topic: {}".format(self.pub_topic))

		def OnSubscribe(self, paho_mqtt, userdata, mid, qos):
			#A new subscription is made
			self.logging.info("Subscribed topic:  {}".format(self.sub_topic))


		def autoStopRT (self):
			'''Timer function that disable automatically the real time streaming
			after a specific time
			'''
			self.logging.debug("Autostop hit")
			self.RTStart.clear()
			self.logging.warning("RT streaming DISABLED")
							




class NewRotatingFileHandler(RotatingFileHandler):

	def __init__(self, *args, **kws):
		super().__init__(*args, **kws)

	def doRollover(self):
		if self.stream:
			self.stream.close()
			self.stream = None
		if self.backupCount > 0:
			for i in range(self.backupCount - 1, 1, -1):
				sfn = self.rotation_filename("%s.%d.gz" % (self.baseFilename, i))
				dfn = self.rotation_filename("%s.%d.gz" % (self.baseFilename, i + 1))
				if os.path.exists(sfn):
					if os.path.exists(dfn):
						os.remove(dfn)
					os.rename(sfn, dfn)
			dfn = self.rotation_filename(self.baseFilename + ".2.gz")
			if os.path.exists(dfn):
				os.remove(dfn)
			

		self.newRotator(self.baseFilename + ".1", dfn)

		sfn = self.rotation_filename(self.baseFilename)
		dfn = self.rotation_filename(self.baseFilename + ".1")
		os.rename(sfn, dfn)
		
		if not self.delay:
			self.stream = self._open()
		

	def newRotator(self,source, dest):
		# Compress the old log.
		if os.path.exists(source):
			with open(source, 'rb') as log:
				with gzip.open(dest, 'wb') as comp_log:
					comp_log.writelines(log)

			os.remove(source)

					
if __name__ == '__main__':

	f_conf=open(confFilename)
	init_conf=json.load(f_conf)
	f_conf.close()

	#initialize the logging module
	handler = NewRotatingFileHandler(filename=logFilename, mode='a', maxBytes=init_conf["log_sizeKB"], backupCount=init_conf["log_count"])
	logging.basicConfig(format='%(asctime)s - %(funcName)s - %(message)s', level=init_conf["log_lev"], handlers=[handler])
	
	
	#take the IMEI of the device from sybok
	ID = 0
	while not ID:
		req_ID = subprocess.Popen("udpctrl -alt property-read F112", stderr=subprocess.PIPE, shell=True)
		ID = req_ID.stderr.read().decode('utf-8').split("=")[1][:-1] 
		ID = ID if len(ID) == 15 else 0
		time.sleep(1)

	#init the can if
	canIf = init_conf["can"]["channel"]

	#init the multiple source tables
	tablesMS = init_conf["can"]["multiple_sources"]

	#init the periodic send interval
	periodic_send = init_conf["mqtt"]["periodic_send"]

	buffer = []
	payload_out = {}
	msg_queue = queue.Queue()

	#threading event to start/stop the CAN processing data
	eventStartRT = threading.Event()
	eventStartCAN = threading.Event()
	bufferLock = threading.Lock()	

	#Initialize the MQTTService class
	ser = MQTTService(init_conf,ID,eventStartRT, logging, version)		
	ser.start()									

	#Thread initialization
	threadCAN = threading.Thread(target=CANFactory, args=(eventStartRT,eventStartCAN, canIf, msg_queue,))
	threadBuffer1 = threading.Thread(target=BufferFactory, args=(msg_queue,eventStartCAN,eventStartRT,bufferLock,))	
	threadBuffer2 = threading.Thread(target=BufferFactory, args=(msg_queue,eventStartCAN,eventStartRT,bufferLock,))
	threadPublisher = threading.Thread(target=periodicPublisher, args=(tablesMS,eventStartRT,eventStartCAN,ser,bufferLock,periodic_send,))
	threadKeyManager = threading.Thread(target=keyOnManager, args=(eventStartCAN,))
	threadCAN.start()
	threadBuffer1.start()
	threadBuffer2.start()
	threadPublisher.start()
	threadKeyManager.start()
	
	#blocking function that wait undefinetely

	while True:

		if not threadCAN.is_alive():
			threadCAN = threading.Thread(target=CANFactory, args=(eventStartRT,eventStartCAN, canIf, msg_queue,))
			threadCAN.start()

		if not threadBuffer1.is_alive():
			threadBuffer1 = threading.Thread(target=BufferFactory, args=(msg_queue,eventStartCAN,eventStartRT,bufferLock,))
			threadBuffer1.start()

		if not threadBuffer2.is_alive():
			threadBuffer2 = threading.Thread(target=BufferFactory, args=(msg_queue,eventStartCAN,eventStartRT,bufferLock,))
			threadBuffer2.start()

		if not threadPublisher.is_alive():
			threadPublisher = threading.Thread(target=periodicPublisher, args=(tablesMS,eventStartRT,eventStartCAN,ser,periodic_send,))
			threadPublisher.start()

		if not threadKeyManager.is_alive():
			threadKeyManager = threading.Thread(target=keyOnManager, args=(eventStartCAN,))
			threadKeyManager.start()

		time.sleep(5)




		



